"""
@ Author      : Troy Kelly
@ Date        : 23 Sept 2018
@ Description : Philips Dynalite Library - Unofficial interface for Philips Dynalite over RS485
"""

from enum import Enum

CONF_ACT_LEVEL = "actual_level"
CONF_ACTION = "action"
CONF_ACTION_CMD = "cmd"
CONF_ACTION_REPORT = "report"
CONF_ACTIVE = "active"
CONF_ACTIVE_ON = "on"
CONF_ACTIVE_OFF = "off"
CONF_ACTIVE_INIT = "init"
CONF_ALL = "ALL"
CONF_AREA = "area"
CONF_CHANNEL = "channel"
CONF_AUTO_DISCOVER = "autodiscover"
CONF_DEFAULT = "default"
CONF_DIR_IN = "IN"
CONF_FADE = "fade"
CONF_HOST = "host"
CONF_JOIN = "join"
CONF_LEVEL = "level"
CONF_LOGLEVEL = "log_level"
CONF_LOGFORMATTER = "log_formatter"
CONF_NAME = "name"
CONF_NODEFAULT = "nodefault"
CONF_PORT = "port"
CONF_POLLTIMER = "polltimer"
CONF_PRESET = "preset"
CONF_STATE = "state"
CONF_STATE_ON = "ON"
CONF_STATE_OFF = "OFF"
CONF_TRGT_LEVEL = "target_level"

EVENT_CHANNEL = "CHANNEL"
EVENT_CONFIGURED = "CONFIGURED"
EVENT_CONNECTED = "CONNECTED"
EVENT_DISCONNECTED = "DISCONNECTED"
EVENT_NEWPRESET = "NEWPRESET"
EVENT_NEWCHANNEL = "NEWCHANNEL"
EVENT_PRESET = "PRESET"
EVENT_REQPRESET = "REQPRESET"

# if a request for channel level didn't get acknowledge, when to retry (subsequent retries will be 2x, 4x, 8x, etc.)
INITIAL_RETRY_DELAY = 1
# initial retry timeout for the beginning, since it could take time to settle on Dynet for large environments
STARTUP_RETRY_DELAY = 60
# Minimal retry frequency in seconds - default 1 hour, so if something never answers, it will be pinged once an hour XXX consider changing
MAXIMUM_RETRY_DELAY = 60 * 60
# no retry value for delay
NO_RETRY_DELAY_VALUE = -1

class SyncType(Enum):
    """Types of Sync Code."""

    LOGICAL = 28  # 0x1c
    DEVICE = 92  # 0x5c
    DEBUG_MSG = 108  # 0x6c

    @classmethod
    def has_value(cls, value):
        """Return the item that has this as a value."""
        return any(value == item.value for item in cls)


class OpcodeType(Enum):
    """Types of Dynet Opcodes."""

    PRESET_1 = 0
    PRESET_2 = 1
    PRESET_3 = 2
    PRESET_4 = 3
    RECALL_OFF = 4
    DECREASE_LEVEL = 5
    INCREMENT_LEVEL = 6
    ENTER_PROGRAM_MODE = 7
    PROGRAM_OUT_CURRENT_PRESET = 8
    PROGRAM_LEVELS_PRESET = 9
    PRESET_5 = 10
    PRESET_6 = 11
    PRESET_7 = 12
    PRESET_8 = 13
    RESET_TO_PRESET = 15
    DMX = 16
    PE_CONTROL = 17
    SET_PE_MIN = 18
    SET_PE_MAX = 19
    AREA_JOIN_LEVEL = 20
    LOCK_CONTROL_PANELS = 21
    ENABLE_CONTROL_PANELS = 22
    PANIC = 23
    UNPANIC = 24
    SET_PE_SPEED = 25
    DISABLE_LIGHT_LEVEL = 26
    ENABLE_LIGHT_LEVEL = 27
    DISABLE_LIGHT_LEVEL_PRESETS = 28
    ENABLE_LIGHT_LEVEL_PRESETS = 29
    DECREMENT_PRESET = 30
    INCREMENT_PRESET = 31
    SET_AREA_LINK = 32
    CLEAR_AREA_LINK = 33
    REPLY_AREA_LINK = 34
    REQUEST_AREA_LINKS = 35
    SET_FADE_TIME = 40
    SUSPEND_OCCUPANCY_PRESETS = 44
    RESUME_OCCUPANCY_PRESETS = 45
    MOTION_ACTIVITY_SYNC = 46
    OCCUPANCY_NOTIFIER = 47
    SUSPEND_OCCUPANCY_SYNC_PRESETS = 48
    SUSPEND_OCCUPANCY_DETECTION_PRESETS = 49
    DISABLE_OCCUPANCY_PRESET = 58
    ENABLE_OCCUPANCY_PRESET = 59
    DISABLE_OCCUPANCY_ALL = 60
    ENABLE_OCCUPANCY_ALL = 61
    AREA_JOIN_MASK = 64
    PANEL_LIGHTING = 72
    REQUEST_AREA_TEMP = 73
    RAMP_ALL_CHANNELS = 95
    REPORT_CHANNEL_LEVEL = 96
    REQUEST_CHANNEL_LEVEL = 97
    REPORT_PRESET = 98
    REQUEST_PRESET = 99
    PRESET_OFFSET = 100
    LINEAR_PRESET = 101
    SAVE_CURRENT_PRESET = 102
    RESTORE_CURRENT_PRESET = 103
    TURN_ALL_AREAS_OFF = 104
    TURN_ALL_AREAS_ON = 105
    FADE_CHANNEL_AREA_TO_PRESET = 107
    TOGGLE_CHANNEL_STATE = 112
    START_FADING_FAST = 113
    START_FADING_MED = 114
    START_FADING_SLOW = 115
    STOP_FADING = 118
    START_FADING_ALL = 121
    STOP_FADING_ALL = 122
    PROGRAM_TOGGLE_PRESET = 125
    SET_CHANNEL_1_TO_LEVEL_WITH_FADE = 128
    SET_CHANNEL_2_TO_LEVEL_WITH_FADE = 129
    SET_CHANNEL_3_TO_LEVEL_WITH_FADE = 130
    SET_CHANNEL_4_TO_LEVEL_WITH_FADE = 131

    @classmethod
    def has_value(cls, value):
        """Return the item that has this as a value."""
        return any(value == item.value for item in cls)



"""
@ Author      : Troy Kelly
@ Date        : 23 Sept 2018
@ Description : Philips Dynalite Library - Unofficial interface for Philips Dynalite over RS485

@ Notes:        Requires a RS485 to IP gateway (Do not use the Dynalite one - use something cheaper)
"""

import asyncio
import logging
import json
import time
from .const import OpcodeType, SyncType, CONF_ACTIVE_ON, CONF_ACTIVE_INIT, CONF_ACTIVE_OFF
from .inbound import DynetInbound

DEFAULT_LOG = logging.getLogger(__name__)


class DynetError(Exception):
    """Class for Dynet errors."""

    def __init__(self, message):
        """Initialize the error."""
        self.message = message


class PacketError(Exception):
    """Class for Dynet packet errors."""

    def __init__(self, message):
        """Initialize the error."""
        self.message = message


class DynetPacket(object):
    """Class for a Dynet network packet."""

    def __init__(self, msg=None, shouldRun=None):
        """Initialize the packet."""
        self.opcodeType = None
        self.sync = None
        self.area = None
        self.data = []
        self.command = None
        self.join = None
        self.chk = None
        if msg is not None:
            self.fromMsg(msg)
        self.shouldRun = shouldRun

    def toMsg(self, sync=28, area=0, command=0, data=[0, 0, 0], join=255):
        """Convert packet to a binary message."""
        bytes = []
        bytes.append(sync)
        bytes.append(area)
        bytes.append(data[0])
        bytes.append(command)
        bytes.append(data[1])
        bytes.append(data[2])
        bytes.append(join)
        bytes.append(self.calcsum(bytes))
        self.fromMsg(bytes)

    def fromMsg(self, msg):
        """Decode a Dynet message."""
        messageLength = len(msg)
        if messageLength < 8:
            raise PacketError("Message too short (%d bytes): %s" % (len(msg), msg))

        if messageLength > 8:
            raise PacketError("Message too long (%d bytes): %s" % (len(msg), msg))

        self._msg = msg

        self.sync = self._msg[0]
        self.area = self._msg[1]
        self.data = [self._msg[2], self._msg[4], self._msg[5]]
        self.command = self._msg[3]
        self.join = self._msg[6]
        self.chk = self._msg[7]
        if self.sync == 28:
            if OpcodeType.has_value(self.command):
                self.opcodeType = OpcodeType(self.command).name

    def toJson(self):
        """Convert to JSON."""
        return json.dumps(self.__dict__)

    def calcsum(self, msg):
        """Calculate the checksum."""
        msg = msg[:7]
        return -(sum(ord(c) for c in "".join(map(chr, msg))) % 256) & 0xFF

    def __repr__(self):
        """Print the packet."""
        return json.dumps(self.__dict__)


class DynetConnection(asyncio.Protocol):
    """Class for an asyncio protocol for the connection to Dynet."""

    def __init__(
        self,
        connectionMade=None,
        connectionLost=None,
        receiveHandler=None,
        connectionPause=None,
        connectionResume=None,
        loop=None,
        logger=DEFAULT_LOG,
    ):
        """Initialize the connection."""
        self._transport = None
        self._paused = False
        self._loop = loop
        self._logger = logger
        self.connectionMade = connectionMade
        self.connectionLost = connectionLost
        self.receiveHandler = receiveHandler
        self.connectionPause = connectionPause
        self.connectionResume = connectionResume

    def connection_made(self, transport):
        """Call when connection is made."""
        self._transport = transport
        self._paused = False
        if self.connectionMade is not None:
            if self._loop is None:
                self.connectionMade(transport)
            else:
                self._loop.create_task(self.connectionMade(transport))

    def connection_lost(self, exc=None):
        """Call when connection is lost."""
        self._transport = None
        if self.connectionLost is not None:
            if self._loop is None:
                self.connectionLost(exc)
            else:
                self._loop.create_task(self.connectionLost(exc))

    def pause_writing(self):
        """Call when connection is paused."""
        self._paused = True
        if self.connectionPause is not None:
            if self._loop is None:
                self.connectionPause()
            else:
                self._loop.create_task(self.connectionPause())

    def resume_writing(self):
        """Call when connection is resumed."""
        self._paused = False
        if self.connectionResume is not None:
            if self._loop is None:
                self.connectionResume()
            else:
                self._loop.create_task(self.connectionResume())

    def data_received(self, data):
        """Call when data is received."""
        if self.receiveHandler is not None:
            if self._loop is None:
                self.receiveHandler(data)
            else:
                self._loop.create_task(self.receiveHandler(data))

    def eof_received(self):
        """Call when EOF for connection."""
        self._logger.debug("EOF Received")


class DynetControl(object):
    """Class to control devices on Dynet network."""

    def __init__(self, dynet, loop, active, areaDefinition=None, logger=DEFAULT_LOG):
        """Initialize the class."""
        self._dynet = dynet
        self._loop = loop
        self.active = active
        self._area = areaDefinition
        self._logger = logger

    def areaPreset(self, area, preset, fade=2):
        """Area preset was set - queue."""
        return self._loop.create_task(
            self._areaPreset(area=area, preset=preset, fade=fade)
        )

    @asyncio.coroutine
    def _areaPreset(self, area, preset, fade):
        """Area preset was set - async."""
        packet = DynetPacket()
        preset = preset - 1
        bank = int((preset) / 8)
        opcode = preset - (bank * 8)
        if opcode > 3:
            opcode = opcode + 6
        fadeLow = int(fade / 0.02) - (int((fade / 0.02) / 256) * 256)
        fadeHigh = int((fade / 0.02) / 256)
        packet.toMsg(
            sync=28, area=area, command=opcode, data=[fadeLow, fadeHigh, bank], join=255
        )
        self._dynet.write(packet)

    def setChannel(self, area, channel, level, fade=2):
        """Set a channel to a given level - queue."""
        return self._loop.create_task(
            self._setChannel(area=area, channel=channel, level=level, fade=fade)
        )

    @asyncio.coroutine
    def _setChannel(self, area, channel, level, fade):
        """Set a channel to a given level - async."""
        packet = DynetPacket()
        channel_bank = 0xFF if (channel <= 4) else (int((channel - 1) / 4) - 1)
        target_level = int(255 - 254 * level)
        opcode = 0x80 + ((channel - 1) % 4)
        fade_time = int(fade / 0.02)
        if (fade_time) > 0xFF:
            fade_time = 0xFF
        packet.toMsg(
            sync=28,
            area=area,
            command=opcode,
            data=[target_level, channel_bank, fade_time],
            join=255,
        )
        self._dynet.write(packet)

    def request_channel_level(self, area, channel, shouldRun=None):
        """Request a level for a specific channel. - queue."""
        return self._loop.create_task(
            self._request_channel_level(area=area, channel=channel, shouldRun=shouldRun)
        )

    @asyncio.coroutine
    def _request_channel_level(self, area, channel, shouldRun):
        """Request a level for a specific channel. - async."""
        packet = DynetPacket(shouldRun=shouldRun)
        packet.toMsg(
            sync=28,
            area=area,
            command=OpcodeType.REQUEST_CHANNEL_LEVEL.value,
            data=[channel - 1, 0, 0],
            join=255,
        )
        self._dynet.write(packet)

    def stop_channel_fade(self, area, channel):
        """Stop fading of a channel - queue."""
        return self._loop.create_task(
            self._stop_channel_fade(area=area, channel=channel)
        )

    @asyncio.coroutine
    def _stop_channel_fade(self, area, channel):
        """Stop fading of a channel - async."""
        packet = DynetPacket()
        packet.toMsg(
            sync=28,
            area=area,
            command=OpcodeType.STOP_FADING.value,
            data=[channel - 1, 0, 0],
            join=255,
        )
        self._dynet.write(packet)

    def areaOff(self, area, fade=2):
        """Turn an area off - queue."""
        return self._loop.create_task(self._areaOff(area=area, fade=fade))

    @asyncio.coroutine
    def _areaOff(self, area, fade):
        """Turn an area off - async."""
        packet = DynetPacket()
        if fade > 25.5:
            fade = 25.5
        if fade < 0:
            fade = 0
        packet.toMsg(
            sync=28, area=area, command=104, data=[255, 0, int(fade * 10)], join=255
        )
        self._dynet.write(packet)

    def request_area_preset(self, area, shouldRun=None):
        """Request current preset of an area - queue."""
        return self._loop.create_task(
            self._request_area_preset(area=area, shouldRun=shouldRun)
        )

    @asyncio.coroutine
    def _request_area_preset(self, area, shouldRun):
        """Request current preset of an area - async."""
        packet = DynetPacket(shouldRun=shouldRun)
        packet.toMsg(
            sync=28,
            area=area,
            command=OpcodeType.REQUEST_PRESET.value,
            data=[0, 0, 0],
            join=255,
        )
        self._dynet.write(packet)


class Dynet(object):
    """Class to handle communication with Dynet network."""

    def __init__(
        self,
        host=None,
        port=None,
        active=CONF_ACTIVE_OFF,
        broadcaster=None,
        onConnect=None,
        onDisconnect=None,
        loop=None,
        logger=DEFAULT_LOG,
    ):
        """Initialize the class."""
        if host is None or port is None or loop is None:
            raise DynetError("Must supply a host, port and loop for Dynet connection")
        self._host = host
        self._port = port
        self._loop = loop
        self._logger = logger
        self.broadcast = broadcaster
        self._onConnect = onConnect
        self._onDisconnect = onDisconnect
        self._conn = lambda: DynetConnection(
            connectionMade=self._connection,
            connectionLost=self._disconnection,
            receiveHandler=self._receive,
            connectionPause=self._pause,
            connectionResume=self._resume,
            loop=self._loop,
        )
        self._transport = None
        self._handlers = {}
        self._connection_retry_timer = 1
        self._paused = False
        self._inBuffer = []
        self._outBuffer = []
        self._timeout = 30
        self.active = active

        self._lastSent = None
        self._messageDelay = 200
        self._sending = False

    def cleanup(self):
        """Clean up with new connection or disconnection."""
        self._connection_retry_timer = 1
        self._transport = None

    def connect(self, onConnect=None):
        """Connect to Dynet - queue."""
        return self._loop.create_task(self._connect())

    async def _connect(self):
        """Connect to Dynet - async."""
        self._logger.debug("Connecting to Dynet on %s:%d" % (self._host, self._port))
        try:
            await asyncio.wait_for(
                self._loop.create_connection(
                    self._conn, host=self._host, port=self._port
                ),
                timeout=self._timeout,
            )
        except (ValueError, OSError, asyncio.TimeoutError) as err:
            self._logger.warning(
                "Could not connect to Dynet (%s). Retrying in %d seconds",
                err,
                self._connection_retry_timer,
            )
            self._loop.call_later(self._connection_retry_timer, self.connect)
            self._connection_retry_timer = (
                2 * self._connection_retry_timer
                if self._connection_retry_timer < 32
                else 60
            )

    @asyncio.coroutine
    def _receive(self, data=None):
        """Handle data that was received."""
        if data is not None:
            for byte in data:
                self._inBuffer.append(int(byte))

        if len(self._inBuffer) < 8:
            self._logger.debug(
                "Received %d bytes, not enough to process: %s"
                % (len(self._inBuffer), self._inBuffer)
            )

        packet = None
        while len(self._inBuffer) >= 8 and packet is None:
            firstByte = self._inBuffer[0]
            if SyncType.has_value(firstByte):
                if firstByte == SyncType.DEBUG_MSG.value:
                    bytemsg = "".join(chr(c) for c in self._inBuffer[1:7])
                    self._logger.debug("Dynet DEBUG message %s" % bytemsg)
                    self._inBuffer = self._inBuffer[8:]
                    continue
                elif firstByte == SyncType.DEVICE.value:
                    self._logger.debug(
                        "Not handling Dynet DEVICE message %s" % self._inBuffer[:8]
                    )
                    self._inBuffer = self._inBuffer[8:]
                    continue
                elif firstByte == SyncType.LOGICAL.value:
                    try:
                        packet = DynetPacket(msg=self._inBuffer[:8])
                    except PacketError as err:
                        self._logger.warning(err)
                        packet = None

            if packet is None:
                hexString = ":".join("{:02x}".format(c) for c in self._inBuffer[:8])
                self._logger.debug(
                    "Unable to process packet %s - moving one byte forward" % hexString
                )
                del self._inBuffer[0]
                continue
            else:
                self._inBuffer = self._inBuffer[8:]

            self._logger.debug("Have packet: %s" % packet)

            if hasattr(packet, "opcodeType") and packet.opcodeType is not None:
                inboundHandler = DynetInbound()
                if hasattr(inboundHandler, packet.opcodeType.lower()):
                    event = getattr(inboundHandler, packet.opcodeType.lower())(packet)
                    if event:
                        self.broadcast(event)
                else:
                    self._logger.debug(
                        "Unhandled Dynet Inbound (%s): %s" % (packet.opcodeType, packet)
                    )
            else:
                self._logger.debug("Unhandled Dynet Inbound: %s" % packet)
        # If there is still buffer to process - start again
        if len(self._inBuffer) >= 8:
            self._loop.create_task(self._receive())

    @asyncio.coroutine
    def _pause(self):
        """Pause transmission on Dynet."""
        self._logger.debug("Pausing Dynet on %s:%d" % (self._host, self._port))
        # Need to schedule a resend here
        self._paused = True

    @asyncio.coroutine
    def _resume(self):
        """Resume transmission on Dynet."""
        self._logger.debug("Resuming Dynet on %s:%d" % (self._host, self._port))
        # Need to schedule a resend here
        self._paused = False

    @asyncio.coroutine
    def _connection(self, transport=None):
        """Handle a new successful connection."""
        self._logger.debug("Connected to Dynet on %s:%d" % (self._host, self._port))
        self.cleanup()
        if transport is not None:
            self.write()  # write whatever is queued in the buffer
            self._transport = transport
            if self._onConnect is not None:
                self._loop.create_task(self._onConnect(dynet=self, transport=transport))
        else:
            raise DynetError("Connected but no transport channel provided")

    @asyncio.coroutine
    def _disconnection(self, exc=None):
        """Handle a network disconnection from Dynet."""
        self._logger.debug(
            "Disconnected from Dynet on %s:%d" % (self._host, self._port)
        )
        self.cleanup()
        if self._onDisconnect is not None:
            self._loop.create_task(self._onDisconnect(dynet=self))

        if exc is not None:
            self._logger.warning(exc)

    def write(self, packet=None):
        """Write a packet or trigger write loop - queue."""
        self._loop.create_task(self._write(packet))

    @asyncio.coroutine
    def _write(self, newPacket=None):
        """Write a packet or trigger write loop - async."""
        if newPacket is not None:
            self._outBuffer.append(newPacket)

        if self._transport is None:
            self._logger.debug("_write before transport is ready. queuing")
            return

        if self._paused or self._sending:
            self._logger.debug("Connection busy - queuing packet")
            self._loop.call_later(1, self.write)
            return

        if self._lastSent is None:
            self._lastSent = int(round(time.time() * 1000))

        current_milli_time = int(round(time.time() * 1000))
        elapsed = current_milli_time - self._lastSent
        delay = 0 - (elapsed - self._messageDelay)
        if delay > 0:
            self._loop.call_later(delay / 1000, self.write)
            return

        if len(self._outBuffer) == 0:
            return

        packet = self._outBuffer[0]
        if packet.shouldRun is None or packet.shouldRun():
            self._sending = True
            msg = bytearray()
            msg.append(packet.sync)
            msg.append(packet.area)
            msg.append(packet.data[0])
            msg.append(packet.command)
            msg.append(packet.data[1])
            msg.append(packet.data[2])
            msg.append(packet.join)
            msg.append(packet.chk)
            assert self.active in [CONF_ACTIVE_ON, CONF_ACTIVE_INIT] or packet.command not in [OpcodeType.REQUEST_CHANNEL_LEVEL.value, OpcodeType.REQUEST_PRESET.value]
            self._transport.write(msg)
            self._logger.debug("Dynet Sent: %s" % msg)
            self._lastSent = int(round(time.time() * 1000))
            self._sending = False

        del self._outBuffer[0]
        if len(self._outBuffer) > 0:
            self._loop.call_later(self._messageDelay / 1000, self.write)



"""Class to represent an event on the Dynet network."""
import json


class DynetEvent(object):
    """Class to represent an event on the Dynet network."""

    def __init__(self, eventType=None, message=None, data={}, direction=None):
        """Initialize the event."""
        self.eventType = eventType.upper() if eventType else None
        self.msg = message
        self.data = data
        self.direction = direction

    def toJson(self):
        """Convert to JSON."""
        return json.dumps(self.__dict__)

    def __repr__(self):
        """Print the event."""
        return json.dumps(self.__dict__)



"""
@ Author      : Troy Kelly
@ Date        : 3 Dec 2018
@ Description : Philips Dynalite Library - Unofficial interface for Philips Dynalite over RS485

@ Notes:        Requires a RS485 to IP gateway (Do not use the Dynalite one - use something cheaper)
"""

import asyncio
import logging
from .dynet import Dynet, DynetControl
from .event import DynetEvent

from .const import (
    CONF_HOST,
    CONF_LOGLEVEL,
    CONF_LOGFORMATTER,
    CONF_PORT,
    CONF_DEFAULT,
    CONF_AREA,
    CONF_NAME,
    CONF_FADE,
    CONF_STATE,
    CONF_STATE_ON,
    CONF_STATE_OFF,
    CONF_LEVEL,
    CONF_PRESET,
    CONF_AUTO_DISCOVER,
    CONF_POLLTIMER,
    CONF_CHANNEL,
    CONF_NODEFAULT,
    CONF_ACTION,
    CONF_ACTION_REPORT,
    CONF_ACTION_CMD,
    CONF_TRGT_LEVEL,
    CONF_ACT_LEVEL,
    CONF_ALL,
    EVENT_CONNECTED,
    EVENT_DISCONNECTED,
    EVENT_CONFIGURED,
    EVENT_NEWPRESET,
    EVENT_NEWCHANNEL,
    EVENT_PRESET,
    EVENT_CHANNEL,
    STARTUP_RETRY_DELAY,
    INITIAL_RETRY_DELAY,
    MAXIMUM_RETRY_DELAY,
    NO_RETRY_DELAY_VALUE,
    CONF_ACTIVE,
    CONF_ACTIVE_ON,
    CONF_ACTIVE_INIT,
)


class BroadcasterError(Exception):
    """Class to handle errors with Broadcaster object."""

    def __init__(self, message):
        """Initialize the error."""
        self.message = message


class PresetError(Exception):
    """Class to handle errors with a Dynalite preset."""

    def __init__(self, message):
        """Initialize the error."""
        self.message = message


class ChannelError(Exception):
    """Class to handle errors with a Dynalite channel."""

    def __init__(self, message):
        """Initialize the error."""
        self.message = message


class AreaError(Exception):
    """Class to handle errors with a Dynalite area."""

    def __init__(self, message):
        """Initialize the error."""
        self.message = message


class DynaliteConfig(object):
    """Class for the configuration of the Dynalite network."""

    def __init__(self, config=None):
        """Initialize the configuration from Dict."""
        self.log_level = (
            config[CONF_LOGLEVEL].upper() if CONF_LOGLEVEL in config else logging.INFO
        )
        self.log_formatter = (
            config[CONF_LOGFORMATTER]
            if CONF_LOGFORMATTER in config
            else "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
        )
        self.host = config[CONF_HOST] if CONF_HOST in config else "localhost"
        self.port = config[CONF_PORT] if CONF_PORT in config else 12345
        self.default = config[CONF_DEFAULT] if CONF_DEFAULT in config else {}
        self.area = config[CONF_AREA] if CONF_AREA in config else {}
        self.preset = config[CONF_PRESET] if CONF_PRESET in config else {}
        self.autodiscover = (
            config[CONF_AUTO_DISCOVER] if CONF_AUTO_DISCOVER in config else False
        )  # autodiscover by default
        self.polltimer = (
            config[CONF_POLLTIMER] if CONF_POLLTIMER in config else 1
        )  # default poll 1 sec
        self.active = config[CONF_ACTIVE] if CONF_ACTIVE in config else False


class Broadcaster(object):
    """Class to broadcast event to listeners."""

    def __init__(self, listenerFunction=None, loop=None, logger=None):
        """Initialize the broadcaster."""
        if listenerFunction is None:
            raise BroadcasterError("A broadcaster bust have a listener Function")
        self._listenerFunction = listenerFunction
        self._monitoredEvents = []
        self._loop = loop
        self.logger = logger

    def monitorEvent(self, eventType=None):
        """Set broadcaster to monitor an event or all."""
        if eventType is None:
            raise BroadcasterError("Must supply an event type to monitor")
        eventType = eventType.upper()
        if eventType not in self._monitoredEvents:
            self._monitoredEvents.append(eventType.upper())

    def unmonitorEvent(self, eventType=None):
        """Stop monitoring an event."""
        if eventType is None:
            raise BroadcasterError("Must supply an event type to un-monitor")
        eventType = eventType.upper()
        if eventType in self._monitoredEvents:
            self._monitoredEvents.remove(eventType.upper())

    def update(self, event=None, dynalite=None):
        """Update listener with an event if relevant."""
        if event is None:
            return
        if (
            event.eventType not in self._monitoredEvents
            and "*" not in self._monitoredEvents
        ):
            return
        if self._loop:
            self._loop.create_task(self._callUpdater(event=event, dynalite=dynalite))
        else:
            self._listenerFunction(event=event, dynalite=dynalite)

    @asyncio.coroutine
    def _callUpdater(self, event=None, dynalite=None):
        """Call listener callback function."""
        self._listenerFunction(event=event, dynalite=dynalite)


class DynalitePreset(object):
    """Class to represent a Dynalite area preset."""

    def __init__(
        self,
        name=None,
        value=None,
        fade=2,
        logger=None,
        broadcastFunction=None,
        area=None,
        dynetControl=None,
    ):
        """Initialize the preset."""
        if not value:
            raise PresetError("A preset must have a value")
        self.logger = logger
        self.active = False
        self.name = name if name else "Preset " + str(value)
        self.value = int(value)
        self.fade = float(fade)
        self.area = area
        self.broadcastFunction = broadcastFunction
        self._control = dynetControl
        if self.broadcastFunction:
            broadcastData = {
                CONF_AREA: self.area.value,
                CONF_PRESET: self.value,
                CONF_NAME: self.name,
                CONF_STATE: "OFF",
            }
            self.broadcastFunction(
                DynetEvent(eventType=EVENT_NEWPRESET, data=broadcastData)
            )

    def turnOn(self, sendDynet=True, sendMQTT=True):
        """Turn the preset on."""
        self.active = True
        if self.area:
            self.area.activePreset = self.value
        if sendMQTT and self.broadcastFunction:
            broadcastData = {
                CONF_AREA: self.area.value,
                CONF_PRESET: self.value,
                CONF_NAME: self.name,
                CONF_STATE: CONF_STATE_ON,
            }
            self.broadcastFunction(
                DynetEvent(eventType=EVENT_PRESET, data=broadcastData)
            )
        if sendDynet and self._control:
            self._control.areaPreset(
                area=self.area.value, preset=self.value, fade=self.fade
            )
        for preset in self.area.preset:
            if self.value != preset:
                if self.area.preset[preset].active:
                    self.area.preset[preset].turnOff(sendDynet=False, sendMQTT=True)
        if self._control.active == CONF_ACTIVE_ON:
            self.area.requestAllChannelLevels(delay=INITIAL_RETRY_DELAY, immediate=False)
        else:
            pass # XXX need to move them to default by preset

    def turnOff(self, sendDynet=True, sendMQTT=True):
        """Turn the preset off."""
        self.active = False
        if sendMQTT and self.broadcastFunction:
            broadcastData = {
                CONF_AREA: self.area.value,
                CONF_PRESET: self.value,
                CONF_NAME: self.name,
                CONF_STATE: CONF_STATE_OFF,
            }
            self.broadcastFunction(
                DynetEvent(eventType=EVENT_PRESET, data=broadcastData)
            )
        if sendDynet and self._control:
            self._control.areaOff(
                area=self.area.value, fade=self.fade
            )  # XXX TODO check if this behavior is correct. In general, you select a preset, so there is no real "turn-off"


class DynaliteChannel(object):
    """Class to represent a Dynalite channel."""

    def __init__(
        self,
        name=None,
        value=None,
        fade=2,
        presets=None,
        logger=None,
        broadcastFunction=None,
        area=None,
        dynetControl=None,
    ):
        """Initialize the channel."""
        logger.debug(
            "DynaliteChannel init called area=%s channel=%s fade=%s"
            % (area.value, value, fade)
        )
        if not value:
            raise ChannelError("A channel must have a value")
        self.logger = logger
        self.level = 0
        self.name = name if name else "Channel " + str(value)
        self.value = int(value)
        self.fade = float(fade)
        self.presets = presets
        self.area = area
        self.broadcastFunction = broadcastFunction
        self._control = dynetControl
        if self.broadcastFunction:
            broadcastData = {
                CONF_AREA: self.area.value,
                CONF_CHANNEL: self.value,
                CONF_NAME: self.name,
                CONF_LEVEL: self.level,
            }
            self.broadcastFunction(
                DynetEvent(eventType=EVENT_NEWCHANNEL, data=broadcastData)
            )
        if self._control.active == CONF_ACTIVE_ON:
            self.requestChannelLevel(
                delay=STARTUP_RETRY_DELAY
            )  # ask for the initial level, but don't resend quickly because the network may still be waiting
        elif self._control.active == CONF_ACTIVE_INIT:
            self.requestChannelLevel(
                delay=NO_RETRY_DELAY_VALUE
            )  # ask for the initial level, but don't retry
            
            
    def turnOn(self, brightness=1.0, sendDynet=True, sendMQTT=True):
        """Turn the channel on or set it to a specific brightness level."""
        if sendDynet and self._control:
            self._control.setChannel(
                area=self.area.value,
                channel=self.value,
                level=brightness,
                fade=self.fade,
            )
        if self._control.active:
            self.requestChannelLevel()
        else:
            self.setLevel(brightness)
            if self.broadcastFunction or True:
                broadcastData = {
                    CONF_AREA: self.area.value,
                    CONF_CHANNEL: self.value,
                    CONF_NAME: self.name,
                    CONF_TRGT_LEVEL: 255 - 254.0 * self.level,
                    CONF_ACTION: CONF_ACTION_CMD,
                }
                self.broadcastFunction(
                    DynetEvent(eventType=EVENT_CHANNEL, data=broadcastData)
                )

    def turnOff(self, sendDynet=True, sendMQTT=True):
        """Turn the channel off."""
        self.turnOn(0, sendDynet, sendMQTT)

    def requestChannelLevel(self, delay=None):
        """Request the channel level."""
        if delay:
            self.area.requestChannelLevel(self.value, delay=delay)
        else:
            self.area.requestChannelLevel(self.value)

    def stopFade(self):
        """Stop fading of the channel."""
        self._control.stop_channel_fade(area=self.area.value, channel=self.value)

    def getLevel(self):
        """Get current channel level."""
        return self.level

    def setLevel(self, level):
        """Set current channel level."""
        self.level = level


class RequestCounter:
    """Helper class to ensure that requests to Dynet for current preset or current channel level get retried but there is only one of each running at each time."""

    def __init__(self, loop, logger=None):
        """Initialize the class."""
        self.loop = loop
        self.logger = logger
        self.counter = 0
        self.timer = None

    def update(self):
        """Update that a new value arrive, so current requests can be cancelled."""
        if self.timer:
            self.timer.cancel()
            self.timer = None
        self.counter += 1

    def timerCallback(self, counter, delay, func, *args):
        """Send request if update was not yet received."""
        self.timer = None
        if self.counter > counter:  # already updated after the timer was scheduled
            return
        func(*args)
        newDelay = min(delay * 2, MAXIMUM_RETRY_DELAY)
        self.timer = self.loop.call_later(
            delay, self.timerCallback, self.counter, newDelay, func, *args
        )

    def schedule(self, delay, immediate, func, *args):
        """Schedule a request until an update arrives with an initial delay and either immediate or not."""
        if delay == NO_RETRY_DELAY_VALUE:
            func(*args)
            return
        if self.timer:
            self.timer.cancel()
        if immediate:
            self.timerCallback(self.counter, delay, func, *args)
        else:
            newDelay = min(delay * 2, MAXIMUM_RETRY_DELAY)
            self.timer = self.loop.call_later(
                delay, self.timerCallback, self.counter, newDelay, func, *args
            )


class DynaliteArea(object):
    """Class to represent a Dynalite area."""

    def __init__(
        self,
        name=None,
        value=None,
        fade=2,
        areaPresets=None,
        defaultPresets=None,
        areaChannels=None,
        areaType=None,
        onPreset=None,
        offPreset=None,
        loop=None,
        logger=None,
        broadcastFunction=None,
        dynetControl=None,
    ):
        """Initialize the area."""
        if not value:
            raise PresetError("An area must have a value")
        self.loop = loop
        self.logger = logger
        self.name = name if name else "Area " + str(value)
        self.type = areaType.lower() if areaType else "light"
        self.value = int(value)
        self.fade = fade
        self.preset = {}
        self.channel = {}
        self.channelUpdateCounter = {}
        self.presetUpdateCounter = RequestCounter(self.loop, self.logger)
        self.activePreset = None
        self.state = None

        if self.type == "cover":
            self._onName = "OPEN"
            self._offName = "CLOSED"
            if onPreset is not None:
                self.openPreset = onPreset
            if offPreset is not None:
                self.closePreset = offPreset
        else:
            self._onName = CONF_STATE_ON
            self._offName = CONF_STATE_OFF
            if onPreset is not None:
                self.onPreset = onPreset
            if offPreset is not None:
                self.offPreset = offPreset

        self.broadcastFunction = broadcastFunction
        self._dynetControl = dynetControl
        
        if self._dynetControl.active == CONF_ACTIVE_ON:
            self.requestPreset(delay=STARTUP_RETRY_DELAY)  # ask for the initial preset
        elif self._dynetControl.active == CONF_ACTIVE_INIT:
            self.requestPreset(delay=NO_RETRY_DELAY_VALUE)  # no retry
        
        if areaPresets:
            for presetValue in areaPresets:
                preset = areaPresets[presetValue]
                presetName = preset[CONF_NAME] if CONF_NAME in preset else None
                presetFade = preset[CONF_FADE] if CONF_FADE in preset else fade
                self.preset[int(presetValue)] = DynalitePreset(
                    name=presetName,
                    value=presetValue,
                    fade=presetFade,
                    logger=self.logger,
                    broadcastFunction=self.broadcastFunction,
                    area=self,
                    dynetControl=self._dynetControl,
                )
        if defaultPresets:
            for presetValue in defaultPresets:
                if int(presetValue) not in self.preset:
                    preset = defaultPresets[presetValue]
                    presetName = preset[CONF_NAME] if preset[CONF_NAME] else None
                    presetFade = preset[CONF_FADE] if preset[CONF_FADE] else fade
                    self.preset[int(presetValue)] = DynalitePreset(
                        name=presetName,
                        value=presetValue,
                        fade=presetFade,
                        logger=self.logger,
                        broadcastFunction=self.broadcastFunction,
                        area=self,
                        dynetControl=self._dynetControl,
                    )
        if areaChannels:
            for channelValue in areaChannels:
                if (int(channelValue) >= 1) and (int(channelValue) <= 255):
                    channel = areaChannels[channelValue]
                    specific_name = (
                        channel[CONF_NAME]
                        if channel and (CONF_NAME in channel)
                        else "Channel " + channelValue
                    )
                    name = self.name + " " + specific_name
                    fade = (
                        channel[CONF_FADE]
                        if channel and (CONF_FADE in channel)
                        else self.fade
                    )  # if no fade provided, use the fade of the area
                    self.channel[int(channelValue)] = DynaliteChannel(
                        name=name,
                        value=channelValue,
                        fade=fade,
                        logger=self.logger,
                        broadcastFunction=self.broadcastFunction,
                        area=self,
                        dynetControl=self._dynetControl,
                    )
                    self.logger.debug(
                        "added area %s channel %s name %s"
                        % (self.name, channelValue, name)
                    )
                else:
                    self.logger.error(
                        "illegal channel value area %s channel %s"
                        % (self.name, channelValue)
                    )
        else:
            self.channel = {}

    def presetOn(self, preset, sendDynet=True, sendMQTT=True, autodiscover=False):
        """Turn a selected preset on and everyone else off."""
        if hasattr(self, "onPreset"):
            if self.onPreset == preset:
                self.state = self._onName
        else:
            self.state = self._offName
        if preset not in self.preset:
            if not autodiscover:
                return
            self.preset[preset] = DynalitePreset(
                value=preset,
                fade=self.fade,
                logger=self.logger,
                broadcastFunction=self.broadcastFunction,
                area=self,
                dynetControl=self._dynetControl,
            )
        self.preset[preset].turnOn(sendDynet=sendDynet, sendMQTT=sendMQTT)

    def presetOff(
        self, preset, sendDynet=True, sendMQTT=True
    ):  # XXX TODO check if this is used anywhere. generally presets cannot be turned off
        """Turn a selected preset off."""
        if preset not in self.preset:
            return  # XXX if we want it to auto-register presets need to fix race condition
            self.preset[preset] = DynalitePreset(
                value=preset,
                fade=self.fade,
                logger=self.logger,
                broadcastFunction=self.broadcastFunction,
                area=self,
                dynetControl=self._dynetControl,
            )
        else:  # this still has a bug that the preset won't be selected when chosen. Need a better fix for the race XXX
            self.preset[preset].turnOff(sendDynet=sendDynet, sendMQTT=sendMQTT)

    def requestPreset(self, delay=INITIAL_RETRY_DELAY, immediate=True):
        """Request the preset for the area."""

        def shouldRun():
            """Return whether or not command is still relevant."""
            return self.presetUpdateCounter.counter == currentCounter

        currentCounter = self.presetUpdateCounter.counter
        self.presetUpdateCounter.schedule(
            delay,
            immediate,
            self._dynetControl.request_area_preset,
            self.value,
            shouldRun,
        )

    def setChannelLevel(self, channel, level, autodiscover=False):
        """Set a channel in an area to a given level. Create it if necessary."""
        if channel in self.channelUpdateCounter:
            self.channelUpdateCounter[channel].update()
        if channel not in self.channel:
            if not autodiscover:
                return
            self.channel[channel] = DynaliteChannel(
                value=channel,
                fade=self.fade,
                logger=self.logger,
                broadcastFunction=self.broadcastFunction,
                area=self,
                dynetControl=self._dynetControl,
            )
        self.channel[channel].setLevel(level)

    def requestChannelLevel(self, channel, delay=INITIAL_RETRY_DELAY, immediate=True):
        """Request the level of a specific channel."""

        def shouldRun():
            """Return whether or not command is still relevant."""
            return self.channelUpdateCounter[channel].counter == currentCounter

        if channel not in self.channelUpdateCounter:
            self.channelUpdateCounter[channel] = RequestCounter(self.loop, self.logger)
        currentCounter = self.channelUpdateCounter[channel].counter
        self.channelUpdateCounter[channel].schedule(
            delay,
            immediate,
            self._dynetControl.request_channel_level,
            self.value,
            channel,
            shouldRun,
        )

    def requestAllChannelLevels(self, delay=INITIAL_RETRY_DELAY, immediate=True):
        """Request channel levels for all channels in an area."""
        if self.channel:
            for channel in self.channel:
                self.requestChannelLevel(channel, delay, immediate)


class Dynalite(object):
    """Class to represent the interaction with Dynalite."""

    def __init__(self, config=None, loop=None, logger=None):
        """Initialize the class."""
        self.loop = loop if loop else asyncio.get_event_loop()
        self.logger = logger if logger else logging.getLogger(__name__)
        self._config = DynaliteConfig(config=config)
        logging.basicConfig(
            level=self._config.log_level, format=self._config.log_formatter
        )

        self._configured = False
        self._autodiscover = False
        self._listeners = []

        self.devices = {CONF_AREA: {}}

        self._dynet = None
        self.control = None

    def start(self):
        """Queue request to start the class."""
        self.loop.create_task(self._start())

    @asyncio.coroutine
    def _start(self):
        """Start the class."""
        self._dynet = Dynet(
            host=self._config.host,
            port=self._config.port,
            active=self._config.active,
            loop=self.loop,
            broadcaster=self.processTraffic,
            onConnect=self._connected,
            onDisconnect=self._disconnection,
        )
        self.control = DynetControl(
            self._dynet, self.loop, self._config.active, areaDefinition=self.devices[CONF_AREA]
        )
        self.connect()  # connect asynchronously. not needed to register devices
        if not self._configured:
            self.loop.create_task(self._configure())

    def connect(self):
        """Queue command to connect to Dynet."""
        self.loop.create_task(self._connect())

    @asyncio.coroutine
    def _connect(self):
        """Connect to Dynet."""
        self._dynet.connect()

    @asyncio.coroutine
    def _connected(self, dynet=None, transport=None):
        """Handle a successful connection."""
        self.broadcast(DynetEvent(eventType=EVENT_CONNECTED, data={}))

    @asyncio.coroutine
    def _disconnection(self, dynet=None):
        """Handle a disconnection and try to reconnect."""
        self.broadcast(DynetEvent(eventType=EVENT_DISCONNECTED, data={}))
        yield from asyncio.sleep(1)  # Don't overload the network
        self.connect()

    def processTraffic(self, event):
        """Process an event that arrived from Dynet - queue."""
        self.loop.create_task(self._processTraffic(event))

    @asyncio.coroutine
    def _processTraffic(self, event):
        """Process an event that arrived from Dynet - async."""
        # The logic here is:
        # - new area is created - ask for the current preset
        # - preset selected - turn the preset on but don't send it as a command
        # - new channel is created - ask for the current level
        # - channel update - update the level and if it is fading (actual != target), schedule a timer to ask again
        # - channel set command - request current level (may not be the target because of fade)
        areaValue = event.data[CONF_AREA]
        if areaValue not in self.devices[CONF_AREA]:
            self.logger.debug("Update from unknown area: %s" % event.toJson)
            if self._autodiscover:
                areaName = "Area " + str(areaValue)
                areaFade = (
                    self._config.default[CONF_FADE]
                    if CONF_FADE in self._config.default
                    else 2
                )
                self.devices[CONF_AREA][areaValue] = DynaliteArea(
                    name=areaName,
                    value=areaValue,
                    fade=areaFade,
                    loop=self.loop,
                    logger=self.logger,
                    broadcastFunction=self.broadcast,
                    dynetControl=self.control,
                )
            else:
                return  # No need to do anything if the area is not defined and we do not have autodiscovery
        curArea = self.devices[CONF_AREA][areaValue]

        if event.eventType in [EVENT_NEWPRESET, EVENT_NEWCHANNEL]:
            self.logger.error(
                event.eventType + " in _processTraffic - we should not get here"
            )
        elif event.eventType == EVENT_PRESET:
            curArea.presetOn(
                event.data[CONF_PRESET],
                sendDynet=False,
                sendMQTT=False,
                autodiscover=self._autodiscover,
            )
            curArea.presetUpdateCounter.update()
        elif event.eventType == EVENT_CHANNEL:
            if event.data[CONF_ACTION] == CONF_ACTION_REPORT:
                if self._config.active == CONF_ACTIVE_ON:
                    curArea.setChannelLevel(
                        event.data[CONF_CHANNEL],
                        (255 - event.data[CONF_ACT_LEVEL]) / 254.0,
                        self._autodiscover,
                    )
                    if event.data[CONF_ACT_LEVEL] != event.data[CONF_TRGT_LEVEL]:
                        self.loop.call_later(
                            self._polltimer,
                            curArea.requestChannelLevel,
                            event.data[CONF_CHANNEL],
                        )
                else:
                    curArea.setChannelLevel(
                        event.data[CONF_CHANNEL],
                        (255 - event.data[CONF_TRGT_LEVEL]) / 254.0,
                        self._autodiscover,
                    )
                    
            elif event.data[CONF_ACTION] == CONF_ACTION_CMD:
                target_level = False
                if CONF_PRESET in event.data:
                    try:
                        target_level = curArea.channel[
                            event.data[CONF_CHANNEL]
                        ].presets[str(event.data[CONF_PRESET])]
                    except (KeyError, TypeError):
                        pass
                if CONF_TRGT_LEVEL in event.data:
                    target_level = (255 - event.data[CONF_TRGT_LEVEL]) / 254.0
                if target_level:  # check if this is relevant for "ALL"
                    if event.data[CONF_CHANNEL] == CONF_ALL:
                        raise Exception(
                            "CHANNEL event with ALL and target_level - should never happen"
                        )  # XXX find a better way to handle it
                    curArea.setChannelLevel(
                        event.data[CONF_CHANNEL], target_level, self._autodiscover
                    )
                if self._config.active == CONF_ACTIVE_ON:
                    if event.data[CONF_CHANNEL] == CONF_ALL:
                        curArea.requestAllChannelLevels()
                    else:
                        curArea.requestChannelLevel(event.data[CONF_CHANNEL])
            else:
                self.logger.warning("CHANNEL command unknown cmd: %s" % event.toJson)
        else:
            self.logger.debug("Unknown event type: %s" % event.toJson)
        # First handle, and then broadcast so broadcast receivers have updated device levels and presets
        self.broadcast(event)

    def broadcast(self, event):
        """Broadcast an event to all listeners - queue."""
        self.loop.create_task(self._broadcast(event))

    @asyncio.coroutine
    def _broadcast(self, event):
        """Broadcast an event to all listeners - async."""
        for listenerFunction in self._listeners:
            listenerFunction.update(event=event, dynalite=self)

    @asyncio.coroutine
    def _configure(self):
        """Configure the class from saved config attribute."""
        self._autodiscover = self._config.autodiscover
        self._polltimer = self._config.polltimer
        for areaValue in self._config.area:
            areaName = (
                self._config.area[areaValue][CONF_NAME]
                if CONF_NAME in self._config.area[areaValue]
                else None
            )
            areaPresets = (
                self._config.area[areaValue][CONF_PRESET]
                if CONF_PRESET in self._config.area[areaValue]
                else {}
            )
            areaFade = (
                self._config.area[areaValue][CONF_FADE]
                if CONF_FADE in self._config.area[areaValue]
                else None
            )
            if areaFade is None:
                areaFade = (
                    self._config.default[CONF_FADE]
                    if CONF_FADE in self._config.default
                    else 2
                )
            areaFade = float(areaFade)
            areaChannels = (
                self._config.area[areaValue][CONF_CHANNEL]
                if CONF_CHANNEL in self._config.area[areaValue]
                else None
            )
            if (
                CONF_NODEFAULT in self._config.area[areaValue]
                and self._config.area[areaValue][CONF_NODEFAULT]
            ):
                defaultPresets = None
            else:
                defaultPresets = self._config.preset

            self.logger.debug(
                "Generating Area '%d/%s' with a default fade of %f"
                % (int(areaValue), areaName, areaFade)
            )
            self.devices[CONF_AREA][int(areaValue)] = DynaliteArea(
                name=areaName,
                value=areaValue,
                fade=areaFade,
                areaPresets=areaPresets,
                areaChannels=areaChannels,
                defaultPresets=defaultPresets,
                loop=self.loop,
                logger=self.logger,
                broadcastFunction=self.broadcast,
                dynetControl=self.control,
            )
        self._configured = True
        self.broadcast(DynetEvent(eventType=EVENT_CONFIGURED, data={}))

    def state(self):
        """Create the state for testing - queue."""
        self.loop.create_task(self._state())

    @asyncio.coroutine
    def _state(self):
        """Create the state for testing - async."""
        for areaValue in self.devices[CONF_AREA]:
            area = self.devices[CONF_AREA][areaValue]
            for presetValue in area.preset:
                preset = area.preset[presetValue]
                presetState = "ON" if preset.active else "OFF"
                broadcastData = {
                    CONF_AREA: area.value,
                    CONF_PRESET: preset.value,
                    CONF_NAME: area.name + " " + preset.name,
                    CONF_STATE: presetState,
                }
                self.broadcast(
                    DynetEvent(eventType=EVENT_NEWPRESET, data=broadcastData)
                )
                if preset.active:
                    self.broadcast(
                        DynetEvent(eventType=EVENT_PRESET, data=broadcastData)
                    )
            self.control.areaReqPreset(area.value)

    def addListener(self, listenerFunction=None):
        """Create a new listener to the class."""
        broadcaster = Broadcaster(
            listenerFunction=listenerFunction, loop=self.loop, logger=self.logger
        )
        self._listeners.append(broadcaster)
        return broadcaster
